//--------------------------------------------------------------------------------------
// File:    KBuffer.h
// Author:  Kevin Myers and Louis Bavoil
// Email:   sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#pragma once
#include "SharedDefines.h"
#include "SubPixelStackCounter.h"
#include "SimpleRT.h"

typedef void (KBUFFER_RENDERLOOP_CALLBACK)( ID3D10Device* pd3dDevice, ID3D10EffectPass* pEffectPass );

class KBuffer
{
public:
    virtual void Capture( ID3D10Device* pd3dDevice, KBUFFER_RENDERLOOP_CALLBACK* pRenderLoopCallback ) = 0;

    virtual void Postprocess( ID3D10Device* pd3dDevice ) = 0;
};

class StencilRoutedKBuffer : public KBuffer
{
public:
    StencilRoutedKBuffer( ID3D10Device* pd3dDevice, ID3D10Effect* m_pEffect, int nWidth, int nHeight )
    {
        HRESULT hr;

        m_pSubPixelStackCounter = new SubPixelStackCounter( pd3dDevice, nWidth, nHeight, MSAA_LEVEL, KBUFFER_PASSES );

        m_pFullScreenPass = m_pEffect->GetTechniqueByName( "Clear" )->GetPassByName( "FSQuadNoPS" );
        V( m_pFullScreenPass && m_pFullScreenPass->IsValid() );

        m_pColorClearPass = m_pEffect->GetTechniqueByName( "Clear" )->GetPassByName( "FSQuadClear" );
        V( m_pColorClearPass && m_pColorClearPass->IsValid() );

        m_pGeometryPass = m_pEffect->GetTechniqueByName( "Geometry" )->GetPassByName( "StencilRouting" );
        V( m_pGeometryPass && m_pGeometryPass->IsValid() );

        m_pPostprocessPass = m_pEffect->GetTechniqueByName( "PostProcessing" )->GetPassByName( "StencilRoutingPost" );
        V( m_pPostprocessPass && m_pPostprocessPass->IsValid() );

        m_pKBufferSRV[0] = m_pEffect->GetVariableByName("tKBuffer0")->AsShaderResource();
        V( m_pKBufferSRV[0] && m_pKBufferSRV[0]->IsValid() );

#if KBUFFER_PASSES > 1
        m_pKBufferSRV[1] = m_pEffect->GetVariableByName("tKBuffer1")->AsShaderResource();
        V( m_pKBufferSRV[1] && m_pKBufferSRV[1]->IsValid() );
#endif

        D3D10_TEXTURE2D_DESC texDesc;
        texDesc.Width = nWidth;
        texDesc.Height = nHeight;
        texDesc.ArraySize = 1;
        texDesc.MiscFlags = 0;
        texDesc.MipLevels = 1;
        texDesc.SampleDesc.Count = MSAA_LEVEL;
        texDesc.SampleDesc.Quality = 0;
        texDesc.Format = DXGI_FORMAT_R32G32_UINT;

        for (UINT8 i = 0; i < KBUFFER_PASSES; ++i) {
            m_KBufferRT[i] = new SimpleRT( pd3dDevice, &texDesc );
            m_ppRTVs[i] = *m_KBufferRT[i];
        }
    }

    ~StencilRoutedKBuffer()
    {
        SAFE_DELETE( m_pSubPixelStackCounter );
        for (UINT8 i = 0; i < KBUFFER_PASSES; ++i) {
            SAFE_DELETE( m_KBufferRT[i] );
        }
    }

    virtual void Capture( ID3D10Device* pd3dDevice, KBUFFER_RENDERLOOP_CALLBACK* pRenderLoopCallback )
    {
        // ClearRenderTargetView cannot clear UINT textures to != 0
        pd3dDevice->OMSetRenderTargets( KBUFFER_PASSES, m_ppRTVs, NULL );
        m_pColorClearPass->Apply( 0 );
        pd3dDevice->IASetInputLayout( NULL );
        pd3dDevice->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP );
        pd3dDevice->Draw( 3, 0 );

        // Capture MSAA_LEVEL fragments per pass
        for (UINT8 i = 0; i < KBUFFER_PASSES; ++i) {
            m_pSubPixelStackCounter->BeginStackRender( pd3dDevice, &m_ppRTVs[i], 1, m_pFullScreenPass, i );
            (*pRenderLoopCallback)( pd3dDevice, m_pGeometryPass );
        }
    }

    virtual void Postprocess( ID3D10Device* pd3dDevice )
    {
        for (UINT8 i = 0; i < KBUFFER_PASSES; ++i) {
            m_pKBufferSRV[i]->SetResource( *m_KBufferRT[i] );
        }
        m_pPostprocessPass->Apply( 0 );

        pd3dDevice->Draw( 3, 0 );

        for (UINT8 i = 0; i < KBUFFER_PASSES; ++i) {
            m_pKBufferSRV[i]->SetResource( NULL );
        }
        m_pPostprocessPass->Apply( 0 );
    }

    virtual void QueryOverflows( ID3D10Device* pd3dDevice )
    {
        m_pSubPixelStackCounter->QueryOverflows( pd3dDevice, m_pFullScreenPass );
    }

    virtual UINT GetNumOverflows()
    {
        return m_pSubPixelStackCounter->EstimateNumOverflows( );
    }

protected:
    SimpleRT* m_KBufferRT[KBUFFER_PASSES];
    ID3D10RenderTargetView *m_ppRTVs[KBUFFER_PASSES];
    ID3D10EffectShaderResourceVariable* m_pKBufferSRV[KBUFFER_PASSES];
    SubPixelStackCounter* m_pSubPixelStackCounter;
    ID3D10EffectPass* m_pFullScreenPass;
    ID3D10EffectPass* m_pColorClearPass;
    ID3D10EffectPass* m_pGeometryPass;
    ID3D10EffectPass* m_pPostprocessPass;
};

class DepthPeeledKBuffer : public KBuffer
{
public:
    DepthPeeledKBuffer( ID3D10Device* pd3dDevice, ID3D10Effect* m_pEffect, int nWidth, int nHeight )
        : m_NumPasses(KBUFFER_PASSES * MSAA_LEVEL)
    {
        HRESULT hr;

        m_pColorClearPass = m_pEffect->GetTechniqueByName( "Clear" )->GetPassByName( "FSQuadClear" );
        V( m_pColorClearPass && m_pColorClearPass->IsValid() );

        m_pDepthPeelingPass = m_pEffect->GetTechniqueByName( "Geometry" )->GetPassByName( "DepthPeeling" );
        V( m_pDepthPeelingPass && m_pDepthPeelingPass->IsValid() );

        m_pPostprocessPass = m_pEffect->GetTechniqueByName( "PostProcessing" )->GetPassByName( "DepthPeelingPost" );
        V( m_pPostprocessPass && m_pPostprocessPass->IsValid() );

        m_pLayersSRV = m_pEffect->GetVariableByName("tLayers")->AsShaderResource();
        V( m_pLayersSRV && m_pLayersSRV->IsValid() );

        m_pDepthBufferSRV = m_pEffect->GetVariableByName("tDepthBuffer")->AsShaderResource();
        V( m_pDepthBufferSRV && m_pDepthBufferSRV->IsValid() );

        D3D10_TEXTURE2D_DESC texDesc;
        texDesc.Width = nWidth;
        texDesc.Height = nHeight;
        texDesc.ArraySize = KBUFFER_PASSES * MSAA_LEVEL;
        texDesc.MiscFlags = 0;
        texDesc.MipLevels = 1;
        texDesc.SampleDesc.Count = 1;
        texDesc.SampleDesc.Quality = 0;
        texDesc.Format = DXGI_FORMAT_R32G32_UINT;
        m_LayersRT = new SimpleArrayRT( pd3dDevice, &texDesc );

        texDesc.ArraySize          = 1;
        texDesc.BindFlags          = D3D10_BIND_DEPTH_STENCIL | D3D10_BIND_SHADER_RESOURCE;
        texDesc.CPUAccessFlags     = NULL;
        texDesc.Format             = DXGI_FORMAT_R32_TYPELESS;
        texDesc.Width              = nWidth;
        texDesc.Height             = nHeight;
        texDesc.MipLevels          = 1;
        texDesc.MiscFlags          = NULL;
        texDesc.SampleDesc.Count   = 1;
        texDesc.SampleDesc.Quality = 0;
        texDesc.Usage              = D3D10_USAGE_DEFAULT;
        for (UINT8 i = 0; i < 2; ++i) {
            m_pDepthBuffers[i] = new DepthRT( pd3dDevice, &texDesc );
        }
    }

    ~DepthPeeledKBuffer()
    {
        SAFE_DELETE( m_LayersRT );
        SAFE_DELETE( m_pDepthBuffers[0] );
        SAFE_DELETE( m_pDepthBuffers[1] );
    }

    virtual void Capture( ID3D10Device* pd3dDevice, KBUFFER_RENDERLOOP_CALLBACK* pRenderLoopCallback )
    {
        // ClearRenderTargetView cannot clear UINT textures to != 0
        m_pColorClearPass->Apply( 0 );
        pd3dDevice->IASetInputLayout( NULL );
        pd3dDevice->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP );
        for (UINT8 i = 0; i < KBUFFER_PASSES; ++i) {
            pd3dDevice->OMSetRenderTargets( MSAA_LEVEL, &m_LayersRT->ppRTVs[i*MSAA_LEVEL], NULL );
            pd3dDevice->Draw( 3, 0 );
        }

        // Initialize depth buffer to 0 to not peel the first layer
        pd3dDevice->ClearDepthStencilView( *m_pDepthBuffers[1], D3D10_CLEAR_DEPTH, 0.0, 0 );

        for (UINT8 layer = 0; layer < m_NumPasses; ++layer) {
            UINT8 currId = layer % 2;
            UINT8 prevId = 1 - currId;

            pd3dDevice->ClearDepthStencilView( *m_pDepthBuffers[currId], D3D10_CLEAR_DEPTH, 1.0, 0 );

            pd3dDevice->OMSetRenderTargets( 1, &m_LayersRT->ppRTVs[layer], NULL );
            m_pDepthBufferSRV->SetResource( *m_pDepthBuffers[prevId] );
            m_pDepthPeelingPass->Apply( 0 );

            pd3dDevice->OMSetRenderTargets( 1, &m_LayersRT->ppRTVs[layer], *m_pDepthBuffers[currId] );
            (*pRenderLoopCallback)( pd3dDevice, m_pDepthPeelingPass );
        }
    }

    virtual void Postprocess( ID3D10Device* pd3dDevice )
    {
        m_pLayersSRV->SetResource( *m_LayersRT );
        m_pPostprocessPass->Apply( 0 );

        pd3dDevice->Draw( 3, 0 );

        m_pLayersSRV->SetResource( NULL );
        m_pPostprocessPass->Apply( 0 );
    }

protected:
    UINT8 m_NumPasses;
    SimpleArrayRT* m_LayersRT;
    ID3D10EffectShaderResourceVariable* m_pLayersSRV;
    DepthRT* m_pDepthBuffers[2];
    ID3D10EffectShaderResourceVariable* m_pDepthBufferSRV;
    ID3D10EffectPass* m_pColorClearPass;
    ID3D10EffectPass* m_pDepthPeelingPass;
    ID3D10EffectPass* m_pPostprocessPass;
};
