//--------------------------------------------------------------------------------------
// File:    SimpleRT.h
// Author:  Kevin Myers and Louis Bavoil
// Email:   sdkfeedback@nvidia.com
//
// Helper classes to simplify rendering to textures and binding them to a shader.
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#pragma once

class SimpleRT 
{
public:
    ID3D10Texture2D* pTexture;
    ID3D10RenderTargetView* pRTV;
    ID3D10ShaderResourceView* pSRV;

    SimpleRT( ID3D10Device* pd3dDevice, D3D10_TEXTURE2D_DESC* pTexDesc )
    {
        HRESULT hr;

        //These have to be set to have a render target
        pTexDesc->BindFlags = D3D10_BIND_RENDER_TARGET | D3D10_BIND_SHADER_RESOURCE;
        pTexDesc->Usage = D3D10_USAGE_DEFAULT;
        pTexDesc->CPUAccessFlags = NULL;

        V( pd3dDevice->CreateTexture2D( pTexDesc, NULL, &pTexture ) );
        V( pd3dDevice->CreateShaderResourceView( pTexture, NULL, &pSRV ) );
        V( pd3dDevice->CreateRenderTargetView( pTexture, NULL, &pRTV ) );
    }

    ~SimpleRT()
    {
        SAFE_RELEASE( pTexture );
        SAFE_RELEASE( pRTV );
        SAFE_RELEASE( pSRV );
    }

    operator ID3D10Texture2D*()
    {
        return pTexture;
    }

    operator ID3D10ShaderResourceView*()
    {
        return pSRV;
    }

    operator ID3D10RenderTargetView*()
    {
        return pRTV;
    }
};

class SimpleArrayRT : public SimpleRT
{
public:
    ID3D10RenderTargetView** ppRTVs;
    UINT iArraySize;

    SimpleArrayRT( ID3D10Device* pd3dDevice, D3D10_TEXTURE2D_DESC* pTexDesc )
        : SimpleRT( pd3dDevice, pTexDesc )
    {
        HRESULT hr;

        D3D10_RENDER_TARGET_VIEW_DESC layerViewDesc;
        pRTV->GetDesc( &layerViewDesc );

        iArraySize = pTexDesc->ArraySize;
        ppRTVs = new ID3D10RenderTargetView*[iArraySize];

        if (iArraySize > 1) {
            switch (layerViewDesc.ViewDimension) {
                case D3D10_RTV_DIMENSION_TEXTURE2DARRAY:
                    layerViewDesc.Texture2DArray.ArraySize = 1;
                    for( UINT i = 0; i < iArraySize; ++i )
                    {
                        layerViewDesc.Texture2DArray.FirstArraySlice = i;
                        V( pd3dDevice->CreateRenderTargetView( pTexture, &layerViewDesc, &ppRTVs[i] ) );
                    }
                    break;
                case D3D10_RTV_DIMENSION_TEXTURE2DMSARRAY:
                    layerViewDesc.Texture2DMSArray.ArraySize = 1;
                    for( UINT i = 0; i < iArraySize; ++i )
                    {
                        layerViewDesc.Texture2DMSArray.FirstArraySlice = i;
                        V( pd3dDevice->CreateRenderTargetView( pTexture, &layerViewDesc, &ppRTVs[i] ) );
                    }
                    break;
            }
        } else {
            V( pd3dDevice->CreateRenderTargetView( pTexture, NULL, &ppRTVs[0] ) );
        }
    }

    ~SimpleArrayRT()
    {
        for (UINT i = 0; i < iArraySize; i++) {
            SAFE_RELEASE( ppRTVs[i] );
        }
        SAFE_DELETE(ppRTVs);
    }

    ID3D10RenderTargetView* GetRenderTargetView(int i)
    {
        return ppRTVs[i];
    }
};

class DepthRT 
{
public:
    ID3D10Texture2D* pTexture;
    ID3D10DepthStencilView*	pDSV;
    ID3D10ShaderResourceView* pSRV;

    DepthRT( ID3D10Device* pd3dDevice, D3D10_TEXTURE2D_DESC* pTexDesc )
    {
        HRESULT hr;
        V( pd3dDevice->CreateTexture2D( pTexDesc, NULL, &pTexture ) );

        D3D10_DEPTH_STENCIL_VIEW_DESC dsvDesc;
        dsvDesc.Format = DXGI_FORMAT_D32_FLOAT;
        dsvDesc.ViewDimension = D3D10_DSV_DIMENSION_TEXTURE2D;
        dsvDesc.Texture2D.MipSlice = 0;
        V( pd3dDevice->CreateDepthStencilView( pTexture, &dsvDesc, &pDSV ) );

        D3D10_SHADER_RESOURCE_VIEW_DESC srvDesc;
        srvDesc.Format = DXGI_FORMAT_R32_FLOAT;
        srvDesc.ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2D;
        srvDesc.Texture2D.MostDetailedMip = 0;
        srvDesc.Texture2D.MipLevels = 1;
        V( pd3dDevice->CreateShaderResourceView( pTexture, &srvDesc, &pSRV ) );
    }

    ~DepthRT()
    {
        SAFE_RELEASE( pTexture );
        SAFE_RELEASE( pDSV );
        SAFE_RELEASE( pSRV );
    }

    operator ID3D10Texture2D*()
    {
        return pTexture;
    }

    operator ID3D10DepthStencilView*()
    {
        return pDSV;
    }

    operator ID3D10ShaderResourceView*()
    {
        return pSRV;
    }
};
