//--------------------------------------------------------------------------------------
// File:    StencilRoutedKBuffer.cpp
// Author:  Kevin Myers and Louis Bavoil
// Email:   sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#include "DXUT.h"
#include "DXUTgui.h"
#include "DXUTmisc.h"
#include "DXUTCamera.h"
#include "DXUTSettingsDlg.h"
#include "SDKmisc.h"
#include "SDKmesh.h"
#include "KBuffer.h"

#define ETA0 1.2f // index of refraction
#define COLOR D3DXVECTOR3(0.3f, 0.7f, 0.1f)
#define CONCENTRATION 8.0f

#define START_WIDTH 1024
#define START_HEIGHT 768

// Direct3D 10 resources
ID3DX10Font* g_pFont = NULL;
ID3DX10Sprite* g_pSprite = NULL;
CDXUTDialogResourceManager g_DialogResourceManager; // manager for shared resources of dialogs
CD3DSettingsDlg g_SettingsDlg; // Device settings dialog
CDXUTDialog g_HUD; // dialog for standard controls
CDXUTDialog g_SampleUI; // dialog for sample specific controls
CModelViewerCamera g_Camera;
CDXUTSDKMesh g_Mesh;



ID3D10Effect* g_pEffect = NULL;
ID3D10Device* g_pd3dDevice = NULL;
UINT g_VideoMemorySizeMB = 0;

ID3D10EffectMatrixVariable* g_pWorldView = NULL;
ID3D10EffectMatrixVariable* g_pWorldViewProj = NULL;
ID3D10EffectMatrixVariable* g_pWorldViewIT = NULL;
ID3D10EffectVectorVariable*	g_pFocalLen = NULL;

ID3D10InputLayout* g_pVertexLayout = NULL;
ID3D10Buffer* g_pVertexBuffer = NULL;
ID3D10Buffer* g_pIndexBuffer = NULL;

StencilRoutedKBuffer* g_pStencilRoutedKBuffer = NULL;
DepthPeeledKBuffer*	g_pDepthPeeledKBuffer = NULL;
KBuffer* g_pKBuffer = NULL;
bool g_UseStencilRouting = true;

ID3D10EffectPass* g_RefractionPass = NULL;
ID3D10EffectShaderResourceVariable* g_pNormalsSRV = NULL;
ID3D10EffectShaderResourceVariable* g_pDepthsSRV = NULL;
ID3D10EffectShaderResourceVariable* g_pColorSRV = NULL;
SimpleArrayRT* g_NormalsRT = NULL;
SimpleArrayRT* g_DepthsRT = NULL;
SimpleRT* g_ColorRT = NULL;
ID3D10RenderTargetView *g_ppRTVs[5];
D3DXVECTOR3 g_AbsorptionSpectrum;

float g_WorldHeightOffset = 0.0;
float g_WorldWidthOffset = 0.0;
bool g_DisplayUI = true;
bool g_bRanOutOfMemory = false;

//--------------------------------------------------------------------------------------
void SetEffect(int i)
{
    g_pEffect->GetVariableByName( "g_Refract" )->AsScalar()->SetBool((i == 0));
    g_pEffect->GetVariableByName( "g_AlphaBlend" )->AsScalar()->SetBool((i == 1));
    g_pEffect->GetVariableByName( "g_LayerId" )->AsScalar()->SetInt(i - 2);
}

//--------------------------------------------------------------------------------------
void ResizeKBuffer( UINT nWidth, UINT nHeight )
{
    SAFE_DELETE(g_pStencilRoutedKBuffer);
    SAFE_DELETE(g_pDepthPeeledKBuffer);
    if (g_UseStencilRouting) {
        g_pStencilRoutedKBuffer = new StencilRoutedKBuffer( g_pd3dDevice, g_pEffect, nWidth, nHeight );
        g_pKBuffer = g_pStencilRoutedKBuffer;
    } else {
        g_pDepthPeeledKBuffer = new DepthPeeledKBuffer( g_pd3dDevice, g_pEffect, nWidth, nHeight );
        g_pKBuffer = g_pDepthPeeledKBuffer;
    }
}

//--------------------------------------------------------------------------------------
enum UIControlIDs {
    IDC_TOGGLEFULLSCREEN = 1,
    IDC_TOGGLEREF,
    IDC_CHANGEDEVICE,
    IDC_QUERIES_CHKBOX,
    IDC_QUERIES_TEXT,
    IDC_CHANGEPOSTFX,
    IDC_CHANGETECHNIQUE,
    IDC_OVERFLOW_COUNTER
};

//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    switch( nControlID )
    {
        case IDC_TOGGLEFULLSCREEN:
            DXUTToggleFullScreen();
            break;
        case IDC_TOGGLEREF:
            DXUTToggleREF();
            break;
        case IDC_CHANGEDEVICE:
            g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() );
            break;
        case IDC_CHANGETECHNIQUE:
            g_UseStencilRouting = (g_SampleUI.GetComboBox( IDC_CHANGETECHNIQUE )->GetSelectedIndex() == 0);
            ResizeKBuffer( DXUTGetDXGIBackBufferSurfaceDesc()->Width, DXUTGetDXGIBackBufferSurfaceDesc()->Height );
            break;
        case IDC_CHANGEPOSTFX:
            SetEffect(g_SampleUI.GetComboBox( IDC_CHANGEPOSTFX )->GetSelectedIndex());
            break;
    }
}

//--------------------------------------------------------------------------------------
void InitSampleUI()
{
    WCHAR sz[100];
    g_SampleUI.Init( &g_DialogResourceManager );

    int iX = 15;
    int iY = 0;
    g_SampleUI.AddComboBox( IDC_CHANGETECHNIQUE, iX, iY, 150, 22 );
    g_SampleUI.GetComboBox( IDC_CHANGETECHNIQUE )->AddItem( L"Stencil Routing", NULL );
    g_SampleUI.GetComboBox( IDC_CHANGETECHNIQUE )->AddItem( L"Depth Peeling", NULL );

    g_SampleUI.AddComboBox( IDC_CHANGEPOSTFX, iX, iY += 32, 150, 22 );
    g_SampleUI.GetComboBox( IDC_CHANGEPOSTFX )->AddItem( L"Translucency", NULL );
    g_SampleUI.GetComboBox( IDC_CHANGEPOSTFX )->AddItem( L"Alpha Blending", NULL );
    for (int i = 0; i < N_MAX; ++i) {
        StringCchPrintf( sz, 100, L"Layer %d", i+1);
        g_SampleUI.GetComboBox( IDC_CHANGEPOSTFX )->AddItem( sz, NULL );
    }

    g_SampleUI.AddStatic( IDC_OVERFLOW_COUNTER, NULL, 25, iY += 28, 125, 22 );

    g_SampleUI.SetCallback( OnGUIEvent );
}

//--------------------------------------------------------------------------------------
void InitApp()
{
    g_SettingsDlg.Init( &g_DialogResourceManager );
    g_HUD.Init( &g_DialogResourceManager );
    g_HUD.SetCallback( OnGUIEvent );
    
    int iY = 10;
    g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, iY, 125, 22 );
    g_HUD.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 35, iY += 24, 125, 22, VK_F3 );
    g_HUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 35, iY += 24, 125, 22, VK_F2 );

    InitSampleUI();

    for (int i = 0; i < 3; ++i) {
        g_AbsorptionSpectrum[i] = -log(COLOR[i]) * CONCENTRATION;
    }
}

//--------------------------------------------------------------------------------------
void RenderText()
{
    CDXUTTextHelper txtHelper(g_pFont, g_pSprite, 15);
    txtHelper.Begin();
    txtHelper.SetInsertionPos( 5, 5 );
    txtHelper.SetForegroundColor( D3DXCOLOR( 0.0f, 0.0f, 0.0f, 1.0f ) );
    txtHelper.DrawTextLine( DXUTGetFrameStats(true) );
    txtHelper.DrawTextLine( DXUTGetDeviceStats() );
    txtHelper.End();
}

//--------------------------------------------------------------------------------------
// Reject any D3D10 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
    return true;
}

//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D10 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
    // For the first device created if its a REF device, optionally display a warning dialog box
    static bool s_bFirstTime = true;
    if( s_bFirstTime )
    {
        s_bFirstTime = false;
        if( (DXUT_D3D9_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d9.DeviceType == D3DDEVTYPE_REF) ||
            (DXUT_D3D10_DEVICE == pDeviceSettings->ver && pDeviceSettings->d3d10.DriverType == D3D10_DRIVER_TYPE_REFERENCE) )
        DXUTDisplaySwitchingToREFWarning( pDeviceSettings->ver );
    }

    // Disable VSync
    pDeviceSettings->d3d10.SyncInterval = 0;

    // Disable MSAA for the backbuffer
    pDeviceSettings->d3d10.sd.SampleDesc.Count = 1;
    pDeviceSettings->d3d10.sd.SampleDesc.Quality = 0;

    return true;
}

//--------------------------------------------------------------------------------------
IDXGIAdapter *GetDXGIAdapter()
{
    IDXGIAdapter* pAdapter;
    for (int i = 0; DXUTGetDXGIFactory()->EnumAdapters(i, &pAdapter) != DXGI_ERROR_NOT_FOUND; ++i) {
        if (pAdapter->CheckInterfaceSupport(__uuidof(ID3D10Device), NULL) == S_OK) {
           return pAdapter;
        }
    }
    return NULL;
}

//--------------------------------------------------------------------------------------
UINT GetVideoMemorySizeMB( ID3D10Device* pd3dDevice )
{
    IDXGIAdapter * pAdapter = GetDXGIAdapter();
    if (pAdapter == NULL) return 0;
    DXGI_ADAPTER_DESC desc;
    pAdapter->GetDesc(&desc);
    SAFE_RELEASE(pAdapter);
    return (UINT)desc.DedicatedVideoMemory / (1024*1024);
}

//--------------------------------------------------------------------------------------
// Create any D3D10 resources that aren't dependant on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

    g_pd3dDevice = pd3dDevice;
    g_VideoMemorySizeMB = GetVideoMemorySizeMB( pd3dDevice );

    V_RETURN( g_DialogResourceManager.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( g_SettingsDlg.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( D3DX10CreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, L"Arial", &g_pFont ) );

    WCHAR str[MAX_PATH];
    V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"StencilRoutedKBuffer.fx" ) );
    D3DX10CreateEffectFromFile( str, NULL, NULL, "fx_4_0", 0, NULL, pd3dDevice, NULL, NULL, &g_pEffect, NULL, NULL );
    V( g_pEffect && g_pEffect->IsValid() );

    g_pWorldView = g_pEffect->GetVariableByName("worldView")->AsMatrix();
    g_pWorldViewProj = g_pEffect->GetVariableByName("worldViewProj")->AsMatrix();
    g_pWorldViewIT = g_pEffect->GetVariableByName("worldViewIT")->AsMatrix();
    g_pFocalLen = g_pEffect->GetVariableByName("g_FocalLen")->AsVector();
    g_pEffect->GetVariableByName( "g_SigmaA" )->AsVector()->SetFloatVector( (float *)&g_AbsorptionSpectrum );

    float r0 = (ETA0 - 1.0f) / (ETA0 + 1.0f);
    g_pEffect->GetVariableByName( "g_R0" )->AsScalar()->SetFloat(r0 * r0);
    g_pEffect->GetVariableByName( "g_EtaInverse" )->AsScalar()->SetFloat(1.0f / ETA0);

    SetEffect(0);

    // Define the input layout
    const D3D10_INPUT_ELEMENT_DESC layout[] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
    };
    UINT numElements = sizeof(layout)/sizeof(layout[0]);

    // Create the input layout
    D3D10_PASS_DESC PassDesc;
    ID3D10EffectTechnique* pTech = g_pEffect->GetTechniqueByName("Geometry");
    pTech->GetPassByIndex( 0 )->GetDesc( &PassDesc );
    V_RETURN( pd3dDevice->CreateInputLayout( layout, numElements, PassDesc.pIAInputSignature, PassDesc.IAInputSignatureSize, &g_pVertexLayout ) );

    // Load the mesh
    V_RETURN( g_Mesh.Create( pd3dDevice, L"..\\Media\\dragon.sdkmesh", true ) );
    g_Camera.SetRadius(1.0f, 0.01f);

    g_RefractionPass = g_pEffect->GetTechniqueByName( "PostProcessing" )->GetPassByName( "Refract" );
    V( g_RefractionPass && g_RefractionPass->IsValid() );

    g_pNormalsSRV = g_pEffect->GetVariableByName("tNormals")->AsShaderResource();
    V( g_pNormalsSRV && g_pNormalsSRV->IsValid() );

    g_pDepthsSRV = g_pEffect->GetVariableByName("tDepths")->AsShaderResource();
    V( g_pDepthsSRV && g_pDepthsSRV->IsValid() );

    g_pColorSRV = g_pEffect->GetVariableByName("tColor")->AsShaderResource();
    V( g_pColorSRV && g_pColorSRV->IsValid() );

    return S_OK;
}

//--------------------------------------------------------------------------------------
// Create any D3D10 resources that depend on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

    UINT nNumPixels = pBackBufferSurfaceDesc->Width * pBackBufferSurfaceDesc->Height;
    if ((nNumPixels > 1920*1200 && g_VideoMemorySizeMB < 500) || (nNumPixels > 1024*768 && g_VideoMemorySizeMB < 300))  //Not enough memory to run at this resolution, just quit
    {
        g_bRanOutOfMemory = true;
        PostQuitMessage( 0 );
    }

    V_RETURN( g_DialogResourceManager.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );
    V_RETURN( g_SettingsDlg.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );

    float fAspectRatio = (float)pBackBufferSurfaceDesc->Width / (float)pBackBufferSurfaceDesc->Height;
    float fovy = D3DX_PI/4;
    g_Camera.SetProjParams( fovy, fAspectRatio, ZNEAR, ZFAR );
    g_Camera.SetWindow( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height);

    float vFocalLen[] = {
        1.0f / tanf(fovy * 0.5f) / fAspectRatio,
        1.0f / tanf(fovy * 0.5f)
    };

    g_pFocalLen->SetFloatVector( vFocalLen );

    g_HUD.SetLocation( pBackBufferSurfaceDesc->Width - 170, 0);
    g_HUD.SetSize( 170, 100 );
    g_SampleUI.SetLocation( pBackBufferSurfaceDesc->Width-170, 110 );
    g_SampleUI.SetSize( 170, 100 );

    D3D10_TEXTURE2D_DESC texDesc;
    texDesc.Width = pBackBufferSurfaceDesc->Width;
    texDesc.Height = pBackBufferSurfaceDesc->Height;
    texDesc.MiscFlags = 0;
    texDesc.MipLevels = 1;
    texDesc.SampleDesc.Count = 1;
    texDesc.SampleDesc.Quality = 0;

    // For refraction, store normals of the first two layers
    texDesc.Format = DXGI_FORMAT_R8G8B8A8_SNORM;
    texDesc.ArraySize = 2;
    SAFE_DELETE( g_NormalsRT );
    g_NormalsRT = new SimpleArrayRT( pd3dDevice, &texDesc );

    // For refraction, store depths of first two layers
    texDesc.Format = DXGI_FORMAT_R32_FLOAT;
    texDesc.ArraySize = 2;
    SAFE_DELETE( g_DepthsRT );
    g_DepthsRT = new SimpleArrayRT( pd3dDevice, &texDesc );

    // Color buffer
    texDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
    texDesc.ArraySize = 1;
    SAFE_DELETE( g_ColorRT );
    g_ColorRT = new SimpleRT( pd3dDevice, &texDesc );

    // MRTs
    for (UINT8 i = 0; i < 2; ++i) g_ppRTVs[i] = g_NormalsRT->GetRenderTargetView(i);
    for (UINT8 i = 0; i < 2; ++i) g_ppRTVs[i+2] = g_DepthsRT->GetRenderTargetView(i);
    g_ppRTVs[4] = *g_ColorRT;

    ResizeKBuffer( pBackBufferSurfaceDesc->Width, pBackBufferSurfaceDesc->Height );

    return S_OK;
}

//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{
    g_Camera.FrameMove( fElapsedTime );

    D3DXMATRIX mProj = *g_Camera.GetProjMatrix();
    D3DXMATRIX mView = *g_Camera.GetViewMatrix();
    D3DXMATRIX mWorld1 = *g_Camera.GetWorldMatrix();

    D3DXMATRIX mTrans, mWorld;
    D3DXMatrixTranslation( &mTrans, -g_WorldWidthOffset, g_WorldHeightOffset, 0 );
    D3DXMatrixMultiply ( &mWorld, &mTrans, &mWorld1 );
    
    D3DXMATRIX mViewI, mViewIT, mViewProj, mViewProjI;
    D3DXMATRIX mWorldI, mWorldIT;
    D3DXMATRIX mWorldView, mWorldViewI, mWorldViewIT;
    D3DXMATRIX mWorldViewProj, mWorldViewProjI;
    D3DXMatrixInverse(&mViewI, NULL, &mView);
    D3DXMatrixTranspose(&mViewIT, &mViewI);
    D3DXMatrixMultiply( &mViewProj, &mView, &mProj );
    D3DXMatrixInverse(&mViewProjI, NULL, &mViewProj);
    D3DXMatrixInverse(&mWorldI, NULL, &mWorld);
    D3DXMatrixTranspose(&mWorldIT, &mWorldI);
    D3DXMatrixMultiply( &mWorldView, &mWorld, &mView );
    D3DXMatrixInverse(&mWorldViewI, NULL, &mWorldView);
    D3DXMatrixTranspose(&mWorldViewIT, &mWorldViewI);
    D3DXMatrixMultiply( &mWorldViewProj, &mWorldView, &mProj );
    D3DXMatrixInverse(&mWorldViewProjI, NULL, &mWorldViewProj);

    g_pWorldView->SetMatrix(mWorldView);
    g_pWorldViewProj->SetMatrix(mWorldViewProj);
    g_pWorldViewIT->SetMatrix(mWorldViewIT);
}

//--------------------------------------------------------------------------------------
static void RenderGeometry( ID3D10Device * pd3dDevice, ID3D10EffectPass* pEffectPass )
{
    UINT Strides[1];
    UINT Offsets[1];
    ID3D10Buffer* pVB[1];
    pVB[0] = g_Mesh.GetVB10(0,0);
    Strides[0] = (UINT)g_Mesh.GetVertexStride(0,0);
    Offsets[0] = 0;
    pd3dDevice->IASetVertexBuffers( 0, 1, pVB, Strides, Offsets );
    pd3dDevice->IASetIndexBuffer( g_Mesh.GetIB10(0), g_Mesh.GetIBFormat10(0), 0 );
    pd3dDevice->IASetInputLayout( g_pVertexLayout );

    pEffectPass->Apply(0);

    for( UINT subset = 0; subset < g_Mesh.GetNumSubsets(0); ++subset )
    {
        SDKMESH_SUBSET* pSubset = g_Mesh.GetSubset( 0, subset );
        D3D10_PRIMITIVE_TOPOLOGY PrimType = g_Mesh.GetPrimitiveType10( (SDKMESH_PRIMITIVE_TYPE)pSubset->PrimitiveType );
        pd3dDevice->IASetPrimitiveTopology( PrimType );
        pd3dDevice->DrawIndexed( (UINT)pSubset->IndexCount, 0, (UINT)pSubset->VertexStart );
    }
}

//--------------------------------------------------------------------------------------
void Refract( ID3D10Device* pd3dDevice )
{
    g_pNormalsSRV->SetResource( *g_NormalsRT );
    g_pDepthsSRV->SetResource( *g_DepthsRT );
    g_pColorSRV->SetResource( *g_ColorRT );
    g_RefractionPass->Apply( 0 );

    pd3dDevice->Draw( 3, 0 );

    g_pNormalsSRV->SetResource( NULL );
    g_pDepthsSRV->SetResource( NULL );
    g_pColorSRV->SetResource( NULL );
    g_RefractionPass->Apply( 0 );
}

//--------------------------------------------------------------------------------------
void GetNumOverflows( ID3D10Device* pd3dDevice )
{
    WCHAR sz[100];
    if (g_UseStencilRouting) {
        g_pStencilRoutedKBuffer->QueryOverflows( pd3dDevice );
        StringCchPrintf( sz, 100, L"Overflowing Pixels: %d", g_pStencilRoutedKBuffer->GetNumOverflows()); 
        g_SampleUI.GetStatic( IDC_OVERFLOW_COUNTER )->SetText( sz );
    } else {
        g_SampleUI.GetStatic( IDC_OVERFLOW_COUNTER )->SetText( L"" );
    }
}

//--------------------------------------------------------------------------------------
// Render the scene using the D3D10 device
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10FrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
    // Clear render target and the depth stencil 
    float ClearColor[4] = { 0.0f, 1.0f, 0.0f, 0.0f };
    pd3dDevice->ClearRenderTargetView( DXUTGetD3D10RenderTargetView(), ClearColor );
    pd3dDevice->ClearDepthStencilView( DXUTGetD3D10DepthStencilView(), D3D10_CLEAR_DEPTH, 1.0, 0 );
    ID3D10RenderTargetView* pBB = DXUTGetD3D10RenderTargetView();
    // If the settings dialog is being shown, then render it instead of rendering the app's scene
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.OnRender( fElapsedTime );
        return;
    }

    g_pKBuffer->Capture( pd3dDevice, RenderGeometry );
    GetNumOverflows( pd3dDevice );

    pd3dDevice->OMSetRenderTargets( 5, g_ppRTVs, NULL );
    g_pKBuffer->Postprocess( pd3dDevice );

    pd3dDevice->OMSetRenderTargets( 1, &pBB, DXUTGetD3D10DepthStencilView() );
    Refract( pd3dDevice );

    if (g_DisplayUI) {
        DXUT_BeginPerfEvent( DXUT_PERFEVENTCOLOR, L"HUD / Stats" );

        RenderText();
        g_HUD.OnRender( fElapsedTime ); 
        g_SampleUI.OnRender( fElapsedTime );

        DXUT_EndPerfEvent();
    }
}

//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10ResizedSwapChain 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10ReleasingSwapChain( void* pUserContext )
{
    g_DialogResourceManager.OnD3D10ReleasingSwapChain();
}

//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10CreateDevice 
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10DestroyDevice( void* pUserContext )
{
    SAFE_DELETE( g_pStencilRoutedKBuffer );
    SAFE_DELETE( g_pDepthPeeledKBuffer );
    SAFE_DELETE( g_ColorRT );
    SAFE_DELETE( g_NormalsRT );
    SAFE_DELETE( g_DepthsRT );
    g_DialogResourceManager.OnD3D10DestroyDevice();
    g_SettingsDlg.OnD3D10DestroyDevice();
    SAFE_RELEASE( g_pFont );
    SAFE_RELEASE( g_pEffect );
    SAFE_RELEASE( g_pVertexLayout );
    SAFE_RELEASE( g_pVertexBuffer );
    SAFE_RELEASE( g_pIndexBuffer );
    g_Mesh.Destroy();
}

//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, 
                          bool* pbNoFurtherProcessing, void* pUserContext )
{
    // Pass messages to dialog resource manager calls so GUI state is updated correctly
    *pbNoFurtherProcessing = g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    // Pass messages to settings dialog if its active
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

    // Give the dialogs a chance to handle the message first
    *pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    *pbNoFurtherProcessing = g_SampleUI.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    g_Camera.HandleMessages( hWnd, uMsg, wParam, lParam );



    return 0;
}

//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
    if(bKeyDown) {
        switch(nChar)
        {
        case 'U':
            g_DisplayUI = !g_DisplayUI;
            break;
        case 'A':
            g_WorldWidthOffset += 0.01f;
            break;
        case 'D':
            g_WorldWidthOffset -= 0.01f;
            break;
        case 'W':
            g_WorldHeightOffset += 0.01f;
            break;
        case 'S':
            g_WorldHeightOffset -= 0.01f;
            break;
        case VK_ESCAPE:
           PostQuitMessage( 0 );
           break;
        }
    }
}

//--------------------------------------------------------------------------------------
// Handle mouse button presses
//--------------------------------------------------------------------------------------
void CALLBACK OnMouse( bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown, 
                       bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta, 
                       int xPos, int yPos, void* pUserContext )
{
}

//--------------------------------------------------------------------------------------
// Call if device was removed.  Return true to find a new device, false to quit
//--------------------------------------------------------------------------------------
bool CALLBACK OnDeviceRemoved( void* pUserContext )
{
    return true;
}

//--------------------------------------------------------------------------------------
// Initialize everything and go into a render loop
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    // DXUT will create and use the best device (either D3D9 or D3D10) 
    // that is available on the system depending on which D3D callbacks are set below

    // Set general DXUT callbacks
    DXUTSetCallbackFrameMove( OnFrameMove );
    DXUTSetCallbackKeyboard( OnKeyboard );
    DXUTSetCallbackMouse( OnMouse );
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );
    DXUTSetCallbackDeviceRemoved( OnDeviceRemoved );

    // Set the D3D10 DXUT callbacks. Remove these sets if the app doesn't need to support D3D10
    DXUTSetCallbackD3D10DeviceAcceptable( IsD3D10DeviceAcceptable );
    DXUTSetCallbackD3D10DeviceCreated( OnD3D10CreateDevice );
    DXUTSetCallbackD3D10SwapChainResized( OnD3D10ResizedSwapChain );
    DXUTSetCallbackD3D10FrameRender( OnD3D10FrameRender );
    DXUTSetCallbackD3D10SwapChainReleasing( OnD3D10ReleasingSwapChain );
    DXUTSetCallbackD3D10DeviceDestroyed( OnD3D10DestroyDevice );

    // IMPORTANT: set SDK media search path to include source directory of this sample, when started from .\Bin
    HRESULT hr;
    V_RETURN(DXUTSetMediaSearchPath(L"..\\Source\\StencilRoutedKBuffer"));
    
    InitApp();
    
    DXUTInit( true, true, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
    DXUTSetCursorSettings( true, true ); // Show the cursor and clip it when in full screen
    DXUTCreateWindow( L"Stencil Routed KBuffer" );
    DXUTCreateDevice( true, START_WIDTH, START_HEIGHT );
    DXUTMainLoop(); // Enter into the DXUT render loop

    // Perform any application-level cleanup here

    if( g_bRanOutOfMemory )  //Post error message for running out of memory
    {
        if( !DXUTGetDeviceSettings().d3d10.sd.Windowed )
            DXUTToggleFullScreen();

        static WCHAR buffer[256];
        StringCchPrintf(buffer, 256, L"Not enough video memory available for this resolution\n");
        
        MessageBox(NULL, buffer, L"Error", MB_OK|MB_ICONERROR);
    }

    return DXUTGetExitCode();
}
