//--------------------------------------------------------------------------------------
// File:    StencilRoutedKBuffer.fx
// Author:  Louis Bavoil and Kevin Myers
// Email:   sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#include "SharedDefines.h"
#define MAX_DEPTH 0xffffffff

Texture2DMS<uint2, MSAA_LEVEL> tKBuffer0;
#if KBUFFER_PASSES > 1
Texture2DMS<uint2, MSAA_LEVEL> tKBuffer1;
#endif
Texture2DArray<uint2> tLayers;
Texture2DArray<float3> tNormals;
Texture2DArray<float> tDepths;
Texture2D<float4> tColor;
Texture2D<float> tDepthBuffer;

cbuffer changingeveryframe
{
    float4x4 worldViewProj;
    float4x4 worldView;
    float4x4 worldViewIT;
};

float3 g_SigmaA;
float2 g_FocalLen;
int g_LayerId;
float g_EtaInverse;
float g_R0;
bool g_Refract;
bool g_AlphaBlend;

//--------------------------------------------------------------------------------------

SamplerState sam
{
    Filter   = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
};

BlendState NoMRT_BS
{
    RenderTargetWriteMask[0] = 0x0f;
};

BlendState MRT2_BS
{
    RenderTargetWriteMask[0] = 0x0f;
    RenderTargetWriteMask[1] = 0x0f;
};

BlendState MRT5_BS
{
    RenderTargetWriteMask[0] = 0x0f;
    RenderTargetWriteMask[1] = 0x0f;
    RenderTargetWriteMask[2] = 0x0f;
    RenderTargetWriteMask[3] = 0x0f;
    RenderTargetWriteMask[4] = 0x0f;
};

DepthStencilState DepthNoStencil_DS
{
    DepthEnable = true;
    DepthFunc = Less_Equal;
    DepthWriteMask = All;
    StencilEnable = false;
};

DepthStencilState NoDepthNoStencil_DS
{
    DepthEnable = false;
    StencilEnable = false;
};

RasterizerState NoCullNoMSAA_RS
{
    CullMode = None;
    FillMode = Solid;
    MultisampleEnable = false;
};

struct QuadOutput
{
    float4 pos : SV_Position;
    float2 tex : TEXCOORD;
};

struct DepthColorOutput
{
    uint2 DepthColor : SV_Target0;
};

//--------------------------------------------------------------------------------------

// Vertex shader that generates a full screen quad with texcoords from vertexIDs
// To use, draw 3 vertices with primitive type triangle strip
QuadOutput FullScreenTriVS( uint id : SV_VertexID )
{
    QuadOutput output = (QuadOutput)0.0f;
    output.tex = float2( (id << 1) & 2, id & 2 );
    output.pos = float4( output.tex * float2( 2.0f, -2.0f ) + float2( -1.0f, 1.0f), 0.0f, 1.0f );
    
    return output;
}

struct DepthColorOutputMRT
{
    uint2 DepthColor0 : SV_Target0;
    uint2 DepthColor1 : SV_Target1;
    uint2 DepthColor2 : SV_Target2;
    uint2 DepthColor3 : SV_Target3;
    uint2 DepthColor4 : SV_Target4;
    uint2 DepthColor5 : SV_Target5;
    uint2 DepthColor6 : SV_Target6;
    uint2 DepthColor7 : SV_Target7;
};

DepthColorOutputMRT ClearPS ( QuadOutput IN ) 
{
    DepthColorOutputMRT output;
    output.DepthColor0 = uint2(MAX_DEPTH, 0);
    output.DepthColor1 = uint2(MAX_DEPTH, 0);
    output.DepthColor2 = uint2(MAX_DEPTH, 0);
    output.DepthColor3 = uint2(MAX_DEPTH, 0);
    output.DepthColor4 = uint2(MAX_DEPTH, 0);
    output.DepthColor5 = uint2(MAX_DEPTH, 0);
    output.DepthColor6 = uint2(MAX_DEPTH, 0);
    output.DepthColor7 = uint2(MAX_DEPTH, 0);
    return output;
}

technique10 Clear
{
    // This pass is used to clear the stencil buffer
    pass FSQuadNoPS
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenTriVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( NULL );
    }
    // This pass is used to clear the R32G32_UINT texture array
    pass FSQuadClear
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenTriVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, ClearPS() ) );
        SetRasterizerState( NoCullNoMSAA_RS );
        SetDepthStencilState( NoDepthNoStencil_DS, 0x00000000 );
        SetBlendState( MRT2_BS, float4( 1.0f, 1.0f, 1.0f, 1.0f ), 0xffffffff );
    }
}

//--------------------------------------------------------------------------------------

uint pack_rgba(float4 c)
{
    uint4 i4 = (uint4) (c * 255.0f);
    return i4.r + (i4.g << 8) + (i4.b << 16) + (i4.a << 24);
}

float4 unpack_rgba(uint c)
{
    float4 c4 = float4(c & 0xff,
        (c >> 8) & 0xff,
        (c >> 16) & 0xff,
        (c >> 24) & 0xff);
    return c4 / 255.0f;
}

uint pack_rgb(float3 c)
{
    uint3 i3 = (uint3) (c * 255.0f);
    return i3.r + (i3.g << 8) + (i3.b << 16);
}

float3 unpack_rgb(uint c)
{
    float3 c3 = float3(c & 0xff, (c >> 8) & 0xff, (c >> 16) & 0xff);
    return c3 / 255.0f;
}

uint pack_normal(float3 n)
{
    uint3 i3 = (uint3) (n * 127.0f + 127.0f);
    return i3.r + (i3.g << 8) + (i3.b << 16);
}

float3 unpack_normal(uint n)
{
    float3 n3 = float3(n & 0xff, (n >> 8) & 0xff, (n >> 16) & 0xff);
    return (n3 - 127.0f) / 127.0f;
}

uint pack_eyez(float z)
{
    return (uint) ((z - ZNEAR) / (ZFAR - ZNEAR) * 0xffffffff);
}

float unpack_eyez(uint z)
{
    return ((float)z / 0xffffffff) * (ZFAR - ZNEAR) + ZNEAR;
}

//--------------------------------------------------------------------------------------

struct Geometry_VSIn
{
    float4 position : position;
    float3 normal   : normal;
};

struct Geometry_VSOut
{
    float4 HPosition     : SV_Position;
    float3 Normal        : normal;
};

Geometry_VSOut GeometryVS ( Geometry_VSIn IN )
{
    Geometry_VSOut OUT;
    OUT.HPosition = mul(IN.position, worldViewProj);
    
    float3 eyeNormal = normalize(mul(IN.normal, worldViewIT).xyz);
    float3 eyePos = mul(IN.position, worldView).xyz;
    OUT.Normal = eyeNormal;
    
    return OUT;
}

DepthColorOutput GeometryPS ( Geometry_VSOut IN ) 
{
    DepthColorOutput output = (DepthColorOutput)0.0f;

    output.DepthColor.x = pack_eyez(IN.HPosition.w);
    output.DepthColor.y = pack_normal(IN.Normal);

    return output;
};

DepthColorOutput DepthPeelPS ( Geometry_VSOut IN ) 
{
    DepthColorOutput output = (DepthColorOutput)0.0f;

    float zFront = tDepthBuffer.Load( int3( IN.HPosition.xy, 0 ) ).r;
    if (IN.HPosition.z <= zFront) discard;
    
    output.DepthColor.x = pack_eyez(IN.HPosition.w);
    output.DepthColor.y = pack_normal(IN.Normal);
    
    return output;
};

technique10 Geometry
{
    pass StencilRouting
    {
        SetVertexShader( CompileShader( vs_4_0, GeometryVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, GeometryPS() ) );
        
        // The DepthStencil state is set in SubPixelStackCounter.h
        SetRasterizerState( NoCullNoMSAA_RS );
        SetBlendState( NoMRT_BS, float4( 1.0f, 1.0f, 1.0f, 1.0f ), 0xffffffff ); 
    }
    pass DepthPeeling
    {
        SetVertexShader( CompileShader( vs_4_0, GeometryVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, DepthPeelPS() ) );

        SetRasterizerState( NoCullNoMSAA_RS );
        SetDepthStencilState( DepthNoStencil_DS, 0x00000000 );
        SetBlendState( NoMRT_BS, float4( 1.0f, 1.0f, 1.0f, 1.0f ), 0xffffffff ); 
    }
}

//--------------------------------------------------------------------------------------

void swap(inout uint2 a, inout uint2 b)
{
    uint2 a1 = a;
    a = b;
    b = a1;
}

// http://www.tools-of-computing.com/tc/CS/Sorts/bitonic_sort.htm
// In this code, k selects the bit position that determines whether the pairs of elements
// are to be exchanged into ascending or descending order. Variable j corresponds to the
// distance apart the elements are that are to be compared and conditionally exchanged.
// Variable i goes through all the elements; ixj is the element that is the pair of element i
// (the exclusive-or of i and j, the element whose position differs only in bit position (log2 j)).
// We only compare elements i and ixj if i<ixj. This avoids comparing them twice.
void BitonicSortF2B( inout uint2 a[N_MAX], int n )
{
    int i,j,k;
    [unroll] for (k=2;k<=n;k=2*k) {
     [unroll] for (j=k>>1;j>0;j=j>>1) {
        [unroll] for (i=0;i<n;i++) {
          int ixj=i^j;
          if ((ixj)>i) {
            if ((i&k)==0 && a[i].x>a[ixj].x) swap(a[i],a[ixj]);
            if ((i&k)!=0 && a[i].x<a[ixj].x) swap(a[i],a[ixj]);
          }
        }
      }
    }
}

//--------------------------------------------------------------------------------------

float3 AlphaBlendF2B(uint2 frag[N_MAX], int nLayers)
{
    float3 dst = (float3)1.0f;
    [unroll] for (int i = nLayers-1; i >= 0; i--) {
        float4 src;
        src.rgb = abs(unpack_normal(frag[i].y).z);
        src.a = 0.6;
        dst.rgb = src.a * src.rgb + (1.0 - src.a) * dst.rgb;
    }
    return dst;
}

float2 EyeToTex(float3 eye)
{
    float2 win = g_FocalLen * eye.xy / eye.z;
    win.xy *= float2(0.5, -0.5);
    win.xy += float2(0.5, 0.5);
    return win;
}

float3 TexToEye(float2 texcoords, float eye_z)
{
    texcoords.xy -= float2(0.5, 0.5);
    texcoords.xy /= float2(0.5, -0.5);
    float2 eye_xy = (texcoords.xy / g_FocalLen) * eye_z;
    return float3(eye_xy, eye_z);
}

float FresnelTransmision(float cosTheta)
{
    // Schlick's approximation to Fresnel's equations
    float k5 = pow(1.0 - cosTheta, 5);
    float R = g_R0 + k5 * (1.0 - g_R0);
    return 1.0 - R;
}

float3 Thickness(uint2 frag[N_MAX], int i0)
{
    float thickness = 0.0f;

    // Compute the thickness along the z axis, assuming a closed object
    int front = i0;
    int back = i0+1;
    for (int i = 0; i < (N_MAX-i0)/2; i++) {
        thickness += unpack_eyez(frag[back].x) - unpack_eyez(frag[front].x);
        back += 2;
        front += 2;
    }

    return thickness;
}

struct LayersOutput
{
    float3 Normal0 : SV_Target0;
    float3 Normal1 : SV_Target1;
    float Depth0   : SV_Target2;
    float Depth1   : SV_Target3;
    float3 Color   : SV_Target4;
};

void RefractionPrepass(inout LayersOutput OUT, uint2 frag[N_MAX], int i0)
{
    // Normals of the first two layers
    OUT.Normal0 = unpack_normal(frag[0].y);
    OUT.Normal1 = unpack_normal(frag[1].y);

    // Eye-space depths of the first two layers
    OUT.Depth0 = (frag[0].y == 0) ? 0 : unpack_eyez(frag[0].x);
    OUT.Depth1 = (frag[1].y == 0) ? 0 : unpack_eyez(frag[1].x);

    // Volumetric absorption through the next layers in the z direction
    float thickness = Thickness(frag, i0);
    OUT.Color = exp(-g_SigmaA * thickness);
}

void Postprocess(inout LayersOutput OUT, uint2 frag[N_MAX])
{
    OUT = (LayersOutput)0.0f;
    
    if (g_Refract) {
        RefractionPrepass(OUT, frag, 2);
        return;
    }
    
    if (g_AlphaBlend) {
        OUT.Color = AlphaBlendF2B(frag, N_MAX);
        return;
    }
    
    float3 N = unpack_normal(frag[g_LayerId].y);
    OUT.Color = abs(N.z);
}

LayersOutput StencilRoutingPostPS ( QuadOutput IN )
{
    LayersOutput OUT;

    uint2 frag[N_MAX];
    [unroll] for (int j = 0; j < MSAA_LEVEL; ++j) {
        frag[j] = tKBuffer0.Load( int3( IN.pos.xy, 0 ), j ).xy;
    }
#if KBUFFER_PASSES > 1    
    [unroll] for (int k = 0; k < MSAA_LEVEL; ++k) {
        frag[MSAA_LEVEL + k] = tKBuffer1.Load( int3( IN.pos.xy, 0 ), k ).xy;
    }
#endif
    
    // Sort the fragments by depth, in front-to-back order
    BitonicSortF2B( frag, N_MAX );
    
    Postprocess( OUT, frag );
    return OUT;
}

LayersOutput DepthPeelingPostPS (  QuadOutput IN )
{
    LayersOutput OUT;
    
    uint2 frag[N_MAX];
    [unroll] for( int i = 0; i < N_MAX; ++i ) {
        frag[i] = tLayers.Load( int4( IN.pos.xy, i, 0 ) ).rg;
    }
    
    Postprocess( OUT, frag );
    return OUT;
}

void ApproxRefract(inout float3 color, inout float2 texcoords, inout float thickness)
{
    // Sample the depth and normal of the first layer
    float z = tDepths.SampleLevel( sam, float3(texcoords, 0), 0 );
    float3 N = tNormals.SampleLevel( sam, float3(texcoords, 0), 0 );
    if (z == 0) return;
    
    // Compute eye-space position P and incident direction I
    float3 P = TexToEye(texcoords, z);
    float3 I = normalize(P);

    // Attenuate through the first layer
    float costheta = -dot(I, N);
    color *= FresnelTransmision(costheta);

    // Do not refract in case of total internal reflection
    float costheta2squared = 1 + (costheta*costheta-1)*(g_EtaInverse*g_EtaInverse);
    if (costheta2squared < 0) {
        return;
    }

    // Compute distance to next layer in the view direction
    float z1 = tDepths.SampleLevel( sam, float3(texcoords, 1), 0 );
    if (z1 == 0) return;
    float3 P1 = TexToEye(texcoords, z1);
    float dv = length(P - P1);
    
    // Compute an approximate hit point
    float costheta2 = sqrt(costheta2squared);
    float3 R = g_EtaInverse * I + (g_EtaInverse*costheta-costheta2) * N;
    float3 hit = P + R * dv;

    // Store candidate for refracted position
    float2 texcoords0 = texcoords;
    texcoords = EyeToTex(hit);
    texcoords = saturate(texcoords);

    // Do not refract if thickness delta < 0
    z1 = tDepths.SampleLevel( sam, float3(texcoords, 1), 0 );
    if (z1 < z) texcoords = texcoords0;

    // Sample the second layer at the perturbed texture coordinates
    z = tDepths.SampleLevel( sam, float3(texcoords, 1), 0 );
    N = tNormals.SampleLevel( sam, float3(texcoords, 1), 0 );
    if (z == 0) return;

    float3 P0 = P;
    P = TexToEye(texcoords, z);
    I = normalize(P - P0);
    thickness += length(P - P0);

    // Attenuate through the second layer
    costheta = abs(dot(I, N));
    color *= FresnelTransmision(costheta);
}

float3 RefractPS ( QuadOutput IN ) : SV_Target
{
    if (!g_Refract) {
        return tColor.Load( int4( IN.pos.xy, 0, 0 ) ).xyz;
    }
    
    float3 color = 1.0f;
    float thickness = 0.0;

    // Refract through the first two layers
    float2 texcoords = IN.tex;
    ApproxRefract(color, texcoords, thickness);
    
    // Compute absorption between the first two layers
    color *= exp(-g_SigmaA * thickness);

    // bgcolor contains the non-refracted color from the other layers
    float3 bgcolor = tColor.SampleLevel( sam, float3(texcoords, 0), 0 ).xyz;
    return color * bgcolor;
}

technique10 PostProcessing
{
    pass StencilRoutingPost
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenTriVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, StencilRoutingPostPS() ) );
        SetRasterizerState( NoCullNoMSAA_RS );
        SetDepthStencilState( NoDepthNoStencil_DS, 0x00000000 );
        SetBlendState( MRT5_BS, float4( 1.0f, 1.0f, 1.0f, 1.0f ), 0xffffffff );
    }
    pass DepthPeelingPost
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenTriVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, DepthPeelingPostPS() ) );
        SetRasterizerState( NoCullNoMSAA_RS );
        SetDepthStencilState( NoDepthNoStencil_DS, 0x00000000 );
        SetBlendState( MRT5_BS, float4( 1.0f, 1.0f, 1.0f, 1.0f ), 0xffffffff );    
    }
    pass Refract
    {
        SetVertexShader( CompileShader( vs_4_0, FullScreenTriVS() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, RefractPS() ) );
        SetRasterizerState( NoCullNoMSAA_RS );
        SetDepthStencilState( NoDepthNoStencil_DS, 0x00000000 );
        SetBlendState( NoMRT_BS, float4( 1.0f, 1.0f, 1.0f, 1.0f ), 0xffffffff );
    }
}
