//--------------------------------------------------------------------------------------
// File:    SubPixelStackCounter.h
// Author:  Kevin Myers and Louis Bavoil
// Email:   sdkfeedback@nvidia.com
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#pragma once
#include <d3d10.h>

class SubPixelStackCounter
{
public:
    SubPixelStackCounter( ID3D10Device* pd3dDevice, UINT nWidth, UINT nHeight, UINT8 nDepth, UINT8 nMaxPasses )
    {
        HRESULT hr;

        m_nDepth = nDepth;
        m_nMaxPasses = nMaxPasses;
        m_pPixelCount = nWidth * nHeight;
        m_bUpdateOverflows = true;

        m_pDSV = NULL;
        m_pDepthStencil = NULL;
        m_pDSWrite = NULL;
        m_pDSUse = NULL;
        m_pDSTest = NULL;
        m_pRS = NULL;
        m_pBS = NULL;
        m_pOCQ = NULL;

        D3D10_TEXTURE2D_DESC texDesc;
        texDesc.ArraySize = 1;
        texDesc.BindFlags = D3D10_BIND_DEPTH_STENCIL;
        texDesc.CPUAccessFlags = NULL;
        texDesc.Format = DXGI_FORMAT_D24_UNORM_S8_UINT;
        texDesc.Width = nWidth;
        texDesc.Height = nHeight;
        texDesc.MipLevels = 1;
        texDesc.MiscFlags = NULL;
        texDesc.SampleDesc.Count = nDepth;
        texDesc.SampleDesc.Quality = 0;
        texDesc.Usage = D3D10_USAGE_DEFAULT;
        
        V( pd3dDevice->CreateTexture2D( &texDesc, NULL, &m_pDepthStencil) );

        D3D10_DEPTH_STENCIL_VIEW_DESC dsvDesc;
        dsvDesc.Format = texDesc.Format;
        dsvDesc.ViewDimension = D3D10_DSV_DIMENSION_TEXTURE2DMS;
        
        V( pd3dDevice->CreateDepthStencilView( m_pDepthStencil, &dsvDesc, &m_pDSV) );

        D3D10_DEPTH_STENCIL_DESC dsDesc;
        dsDesc.DepthEnable = false;
        dsDesc.DepthWriteMask = D3D10_DEPTH_WRITE_MASK_ZERO;
        dsDesc.DepthFunc = D3D10_COMPARISON_ALWAYS;
        
        dsDesc.StencilEnable = true;
        dsDesc.StencilReadMask = 0xff;
        dsDesc.StencilWriteMask = 0xff;

        dsDesc.FrontFace.StencilFunc = D3D10_COMPARISON_ALWAYS;
        dsDesc.BackFace.StencilFunc = D3D10_COMPARISON_ALWAYS;

        dsDesc.FrontFace.StencilDepthFailOp = D3D10_STENCIL_OP_KEEP;
        dsDesc.BackFace.StencilDepthFailOp = D3D10_STENCIL_OP_KEEP;
        
        dsDesc.FrontFace.StencilFailOp = D3D10_STENCIL_OP_REPLACE;
        dsDesc.FrontFace.StencilPassOp = D3D10_STENCIL_OP_REPLACE;

        dsDesc.BackFace.StencilFailOp = D3D10_STENCIL_OP_REPLACE;
        dsDesc.BackFace.StencilPassOp = D3D10_STENCIL_OP_REPLACE;
        

        V( pd3dDevice->CreateDepthStencilState( &dsDesc, &m_pDSWrite ) );

        dsDesc.FrontFace.StencilFunc = D3D10_COMPARISON_EQUAL;
        dsDesc.BackFace.StencilFunc = D3D10_COMPARISON_EQUAL;

        dsDesc.BackFace.StencilFailOp = D3D10_STENCIL_OP_DECR_SAT;
        dsDesc.BackFace.StencilPassOp = D3D10_STENCIL_OP_DECR_SAT;

        dsDesc.FrontFace.StencilFailOp = D3D10_STENCIL_OP_DECR_SAT;
        dsDesc.FrontFace.StencilPassOp = D3D10_STENCIL_OP_DECR_SAT;

        V( pd3dDevice->CreateDepthStencilState( &dsDesc, &m_pDSUse ) );

        dsDesc.FrontFace.StencilFunc = D3D10_COMPARISON_EQUAL;
        dsDesc.BackFace.StencilFunc = D3D10_COMPARISON_EQUAL;

        dsDesc.BackFace.StencilFailOp = D3D10_STENCIL_OP_KEEP;
        dsDesc.BackFace.StencilPassOp = D3D10_STENCIL_OP_KEEP;

        dsDesc.FrontFace.StencilFailOp = D3D10_STENCIL_OP_KEEP;
        dsDesc.FrontFace.StencilPassOp = D3D10_STENCIL_OP_KEEP;

        V( pd3dDevice->CreateDepthStencilState( &dsDesc, &m_pDSTest ) );

        D3D10_BLEND_DESC bsDesc;
        ZeroMemory( &bsDesc, sizeof( D3D10_BLEND_DESC ) );
        V( pd3dDevice->CreateBlendState( &bsDesc, &m_pBS ) );

        D3D10_RASTERIZER_DESC rsDesc;
        rsDesc.AntialiasedLineEnable = false;
        rsDesc.DepthBias = 0;
        rsDesc.DepthBiasClamp = 0.0f;
        rsDesc.FrontCounterClockwise = false;
        rsDesc.ScissorEnable = false;
        rsDesc.SlopeScaledDepthBias = 0.0f;

        rsDesc.CullMode = D3D10_CULL_NONE;
        rsDesc.FillMode = D3D10_FILL_SOLID;
        rsDesc.MultisampleEnable = false;

        V( pd3dDevice->CreateRasterizerState( &rsDesc, &m_pRS ) );

        m_VP.TopLeftX = m_VP.TopLeftY = 0;
        m_VP.Width = nWidth;
        m_VP.Height = nHeight;
        m_VP.MinDepth = 0.0f;
        m_VP.MaxDepth = 1.0f;

        D3D10_QUERY_DESC qDesc;
        qDesc.Query = D3D10_QUERY_OCCLUSION;
        qDesc.MiscFlags = 0;
        V( pd3dDevice->CreateQuery( &qDesc, &m_pOCQ ) );
    }

    void BeginStackRender( ID3D10Device* pd3dDevice, ID3D10RenderTargetView** ppRTVs, UINT nRTVs, ID3D10EffectPass* pFullScreenPass, UINT8 nPass )
    {
        pFullScreenPass->Apply( 0 );

        pd3dDevice->IASetInputLayout( NULL );
        pd3dDevice->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP );
        pd3dDevice->RSSetState( m_pRS );

        pd3dDevice->RSSetViewports( 1, &m_VP );
        pd3dDevice->OMSetRenderTargets( 0, NULL, m_pDSV );

        //This fills the stencil buffer with a subpixel mask
        //each subpixel gets a different value
        {
            const float blendFactor[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
            UINT nSampleMask = 1;
            UINT8 nStencilRef = m_nDepth * ( nPass + 1 ) + 1;

            pd3dDevice->ClearDepthStencilView( m_pDSV, D3D10_CLEAR_STENCIL, 1.0f, nStencilRef );  //This clear essentially fils the first sample
            for( UINT8 i = 1; i < m_nDepth; ++i )
            {
                nSampleMask = nSampleMask << 1;
                --nStencilRef;

                pd3dDevice->OMSetBlendState( m_pBS, blendFactor, nSampleMask );
                pd3dDevice->OMSetDepthStencilState( m_pDSWrite, nStencilRef );
                pd3dDevice->Draw( 3, 0 );
            }
        }

        pd3dDevice->OMSetRenderTargets( nRTVs, ppRTVs, m_pDSV );
        pd3dDevice->OMSetDepthStencilState( m_pDSUse, 0x00000002 );
    }

    void QueryOverflows( ID3D10Device* pd3dDevice, ID3D10EffectPass* pFullScreenPass )
    {
        if( m_bUpdateOverflows ) {
            pFullScreenPass->Apply( 0 );

            const float blendFactor[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
            pd3dDevice->IASetInputLayout( NULL );
            pd3dDevice->IASetPrimitiveTopology( D3D10_PRIMITIVE_TOPOLOGY_TRIANGLESTRIP );

            pd3dDevice->RSSetState( m_pRS );
            pd3dDevice->RSSetViewports( 1, &m_VP );

            pd3dDevice->OMSetDepthStencilState( m_pDSTest, 0x00000000 );
            pd3dDevice->OMSetBlendState( m_pBS, blendFactor, 1 );
            pd3dDevice->OMSetRenderTargets( 0, NULL, m_pDSV );

            m_pOCQ->Begin();
            pd3dDevice->Draw( 3, 0 );
            m_pOCQ->End();
        }
    }

    UINT GetNumOverflows( )
    {
        UINT64 nSamples = 0;
        while( m_pOCQ->GetData( &nSamples, sizeof(UINT64), NULL ) != S_OK ){}
        return (UINT)nSamples;
    }

    UINT EstimateNumOverflows( )
    {
        static UINT s_nSamples = 0;
        UINT64 nSamples = 0;
        if ( m_pOCQ->GetData( &nSamples, sizeof(UINT64), NULL ) == S_OK ) {
            m_bUpdateOverflows = true;
            s_nSamples = (UINT)nSamples;
        } else {
            m_bUpdateOverflows = false;
        }
        return s_nSamples;
    }

    ~SubPixelStackCounter()
    {
        SAFE_RELEASE( m_pDepthStencil );
        SAFE_RELEASE( m_pDSV );
        SAFE_RELEASE( m_pDSWrite );
        SAFE_RELEASE( m_pDSUse );
        SAFE_RELEASE( m_pDSTest );
        SAFE_RELEASE( m_pBS );
        SAFE_RELEASE( m_pRS );
        SAFE_RELEASE( m_pOCQ );
    };

private:
    UINT8 m_nDepth;
    UINT8 m_nMaxPasses;
    UINT64 m_pPixelCount;
    D3D10_VIEWPORT m_VP;
    bool m_bUpdateOverflows;
    UINT m_nQueriesIssued;
    ID3D10Query* m_pOCQ;

    ID3D10DepthStencilView* m_pDSV;
    ID3D10Texture2D* m_pDepthStencil;

    ID3D10DepthStencilState* m_pDSWrite;
    ID3D10DepthStencilState* m_pDSUse;
    ID3D10DepthStencilState* m_pDSTest;
    ID3D10RasterizerState* m_pRS;
    ID3D10BlendState* m_pBS;
};
